"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const rollup = tslib_1.__importStar(require("rollup"));
const rollup_plugin_generate_package_json_1 = tslib_1.__importDefault(require("rollup-plugin-generate-package-json"));
const rollup_plugin_auto_external_1 = tslib_1.__importDefault(require("rollup-plugin-auto-external"));
const plugin_node_resolve_1 = tslib_1.__importDefault(require("@rollup/plugin-node-resolve"));
const rollup_plugin_typescript2_1 = tslib_1.__importDefault(require("rollup-plugin-typescript2"));
const globby_1 = tslib_1.__importDefault(require("globby"));
const p_limit_1 = tslib_1.__importDefault(require("p-limit"));
const fs_extra_1 = tslib_1.__importDefault(require("fs-extra"));
const path_1 = require("path");
const lodash_get_1 = tslib_1.__importDefault(require("lodash.get"));
const mkdirp_1 = tslib_1.__importDefault(require("mkdirp"));
const command_1 = require("../command");
const distDir = "dist";
exports.buildCommand = command_1.createCommand((api) => {
    const { config, reporter } = api;
    return {
        command: "build",
        describe: "Build",
        builder(yargs) {
            return yargs.options({
                single: {
                    describe: "Single package (THE OPOSITE OF MONOREPO)",
                    type: "boolean",
                },
            });
        },
        async handler(args) {
            if (args.single) {
                await buildSingle();
                return;
            }
            const limit = p_limit_1.default(4);
            const packages = await globby_1.default("packages/**/package.json", {
                cwd: process.cwd(),
                absolute: true,
                ignore: ["**/node_modules/**", `**/${distDir}/**`],
            });
            await Promise.all(packages.map((packagePath) => limit(() => build(packagePath, config, reporter))));
        },
    };
});
async function buildSingle() {
    const cwd = process.cwd();
    const packagePath = path_1.join(process.cwd(), "package.json");
    const pkg = await readPackageJson(cwd);
    validatePackageJson(pkg);
    const buildOptions = pkg.buildOptions || {};
    const extraInputOptions = {};
    if (buildOptions.external) {
        extraInputOptions.external = buildOptions.external;
    }
    const inputOptions = {
        input: "src/index.ts",
        plugins: [
            plugin_node_resolve_1.default(),
            rollup_plugin_auto_external_1.default({
                packagePath,
                builtins: true,
                dependencies: true,
                peerDependencies: true,
            }),
            rollup_plugin_typescript2_1.default(),
            rollup_plugin_generate_package_json_1.default({
                baseContents: rewritePackageJson(pkg),
                additionalDependencies: Object.keys(pkg.dependencies || {}),
            }),
        ],
        inlineDynamicImports: true,
        ...extraInputOptions,
    };
    // create a bundle
    const bundle = await rollup.rollup(inputOptions);
    // generates
    const commonOutputOptions = {
        preferConst: true,
        sourcemap: true,
    };
    const generates = [
        {
            ...commonOutputOptions,
            file: path_1.join(distDir, "index.cjs.js"),
            format: "cjs",
        },
        {
            ...commonOutputOptions,
            file: path_1.join(distDir, "index.esm.js"),
            format: "esm",
        },
    ];
    if (pkg.exports) {
        generates.push({
            ...commonOutputOptions,
            file: path_1.join(distDir, "index.mjs"),
            format: "esm",
        });
    }
    await Promise.all(generates.map(async (outputOptions) => {
        await bundle.write(outputOptions);
    }));
    // move README.md and LICENSE
    await copyToDist(cwd, ["README.md", "LICENSE"].concat((buildOptions === null || buildOptions === void 0 ? void 0 : buildOptions.copy) || []));
}
async function build(packagePath, config, reporter) {
    var _a;
    const scope = config.scope;
    const cwd = packagePath.replace("/package.json", "");
    const pkg = await readPackageJson(cwd);
    const fullName = pkg.name;
    if ((config.ignore || []).includes(fullName)) {
        reporter.warn(`Ignored ${fullName}`);
        return;
    }
    const name = fullName.replace(`${scope}/`, "");
    validatePackageJson(pkg);
    const distProjectDir = cwd.replace("packages", distDir);
    const distProjectSrcDir = path_1.resolve(distProjectDir, "src");
    const bobDir = path_1.resolve(process.cwd(), ".bob");
    const bobProjectDir = path_1.resolve(bobDir, name);
    const buildOptions = pkg.buildOptions || {};
    // remove bob/<project-name>
    await fs_extra_1.default.remove(bobProjectDir);
    const inputFile = path_1.resolve(distProjectSrcDir, "index.js");
    const extraInputOptions = {};
    if (buildOptions.external) {
        extraInputOptions.external = buildOptions.external;
    }
    const inputOptions = {
        input: inputFile,
        plugins: [
            plugin_node_resolve_1.default(),
            rollup_plugin_auto_external_1.default({
                packagePath,
                builtins: true,
                dependencies: true,
                peerDependencies: true,
            }),
            rollup_plugin_generate_package_json_1.default({
                baseContents: rewritePackageJson(pkg),
                additionalDependencies: Object.keys(pkg.dependencies || {}),
            }),
        ],
        inlineDynamicImports: true,
        ...extraInputOptions,
    };
    // create a bundle
    const bundle = await rollup.rollup(inputOptions);
    // generates
    const commonOutputOptions = {
        preferConst: true,
        sourcemap: true,
    };
    const generates = [
        {
            ...commonOutputOptions,
            file: path_1.join(bobProjectDir, "index.cjs.js"),
            format: "cjs",
        },
        {
            ...commonOutputOptions,
            file: path_1.join(bobProjectDir, "index.esm.js"),
            format: "esm",
        },
    ];
    if (pkg.exports) {
        generates.push({
            ...commonOutputOptions,
            file: path_1.join(distDir, "index.mjs"),
            format: "esm",
        });
    }
    const declarations = await globby_1.default("**/*.d.ts", {
        cwd: distProjectSrcDir,
        ignore: ["**/node_modules/**"],
    });
    const limit = p_limit_1.default(200);
    await Promise.all(generates.map(async (outputOptions) => {
        await bundle.write(outputOptions);
    }));
    await Promise.all(declarations.map((file) => limit(() => fs_extra_1.default.copy(path_1.join(distProjectSrcDir, file), path_1.join(bobProjectDir, file)))));
    if (buildOptions === null || buildOptions === void 0 ? void 0 : buildOptions.bin) {
        await Promise.all(Object.keys(buildOptions.bin).map(async (alias) => {
            const options = buildOptions.bin[alias];
            const binPath = path_1.resolve(distProjectSrcDir, options.input.replace("src/", "").replace(".ts", ".js"));
            const inputOptions = {
                input: binPath,
                plugins: [
                    plugin_node_resolve_1.default(),
                    rollup_plugin_auto_external_1.default({
                        packagePath,
                        builtins: true,
                        dependencies: true,
                        peerDependencies: true,
                    }),
                ],
                inlineDynamicImports: true,
            };
            const bundle = await rollup.rollup(inputOptions);
            await bundle.write({
                banner: `#!/usr/bin/env node`,
                preferConst: true,
                sourcemap: options.sourcemap,
                file: path_1.join(bobProjectDir, pkg.bin[alias].replace(`${distDir}/`, "")),
                format: "cjs",
            });
        }));
    }
    // remove <project>/dist
    await fs_extra_1.default.remove(path_1.join(cwd, distDir));
    // move bob/<project-name> to <project>/dist
    await fs_extra_1.default.move(bobProjectDir, path_1.join(cwd, distDir));
    // move README.md and LICENSE
    await copyToDist(cwd, ["README.md", "LICENSE"].concat(((_a = pkg.buildOptions) === null || _a === void 0 ? void 0 : _a.copy) || []));
    reporter.success(`Built ${pkg.name}`);
}
//
async function readPackageJson(baseDir) {
    return JSON.parse(await fs_extra_1.default.readFile(path_1.resolve(baseDir, "package.json"), {
        encoding: "utf-8",
    }));
}
exports.readPackageJson = readPackageJson;
function rewritePackageJson(pkg) {
    const newPkg = {};
    const fields = [
        "name",
        "version",
        "description",
        "sideEffects",
        "peerDependencies",
        "dependencies",
        "optionalDependencies",
        "repository",
        "homepage",
        "keywords",
        "author",
        "license",
        "engines",
    ];
    fields.forEach((field) => {
        if (typeof pkg[field] !== "undefined") {
            newPkg[field] = pkg[field];
        }
    });
    newPkg.main = "index.cjs.js";
    newPkg.module = "index.esm.js";
    newPkg.typings = "index.d.ts";
    newPkg.typescript = {
        definition: newPkg.typings,
    };
    if (pkg.bin) {
        newPkg.bin = {};
        for (const alias in pkg.bin) {
            newPkg.bin[alias] = pkg.bin[alias].replace(`${distDir}/`, "");
        }
    }
    return newPkg;
}
function validatePackageJson(pkg) {
    function expect(key, expected) {
        const received = lodash_get_1.default(pkg, key);
        if (expected !== received) {
            throw new Error(`${pkg.name}: "${key}" equals "${received}", should be "${expected}"`);
        }
    }
    expect("main", `${distDir}/index.cjs.js`);
    expect("module", `${distDir}/index.esm.js`);
    expect("typings", `${distDir}/index.d.ts`);
    expect("typescript.definition", `${distDir}/index.d.ts`);
    if (pkg.exports) {
        expect("exports.require", pkg.main);
        expect("exports.default", `${distDir}/index.mjs`);
    }
}
exports.validatePackageJson = validatePackageJson;
async function copyToDist(cwd, files) {
    const allFiles = await globby_1.default(files, { cwd });
    return Promise.all(allFiles.map(async (file) => {
        if (await fs_extra_1.default.pathExists(path_1.join(cwd, file))) {
            const sourcePath = path_1.join(cwd, file);
            const destPath = path_1.join(cwd, distDir, file.replace('src/', ''));
            await mkdirp_1.default(path_1.dirname(destPath));
            await fs_extra_1.default.copyFile(sourcePath, destPath);
        }
    }));
}
