"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const util_1 = require("util");
const child_process_1 = require("child_process");
const cross_spawn_1 = tslib_1.__importDefault(require("cross-spawn"));
const minimatch_1 = tslib_1.__importDefault(require("minimatch"));
const command_1 = require("../command");
exports.runCommand = command_1.createCommand((api) => {
    const { config, reporter } = api;
    return {
        command: "run <command>",
        describe: "Runs a command only for affected packages",
        builder(yargs) {
            return yargs.positional("command", {
                describe: "Name of the command",
                type: "string",
                demandOption: true,
            });
        },
        async handler(args) {
            var _a;
            const commandName = args.command;
            const command = (_a = config === null || config === void 0 ? void 0 : config.commands) === null || _a === void 0 ? void 0 : _a[commandName];
            if (!command) {
                // Better error - example config
                throw new Error(`Command ${commandName} not found. Define it in bob.config.js`);
            }
            const { affected, packages } = getAffectedPackages({
                config,
                filterCommand: commandName,
            });
            if (!affected.length) {
                reporter.success("Nothing is affected");
                return;
            }
            reporter.info([
                `Affected packages: `,
                affected.map((name) => ` - ${name}`).join("\n"),
                "\n\n",
            ].join("\n"));
            const input = {
                names: affected,
                paths: affected.map((name) => packages[name].location),
            };
            const [bin, rest] = await command.run(input);
            await util_1.promisify(cross_spawn_1.default)(bin, rest, {
                stdio: "inherit",
            });
        },
    };
});
function getAffectedPackages({ config, filterCommand, }) {
    if (!config.track || config.track.length === 0) {
        throw new Error(`Define files to track`);
    }
    if (!config.base) {
        throw new Error(`Define 'against' in config. In most cases set it to 'origin/master`);
    }
    const packages = getPackages(config.ignore);
    const changedFiles = getChangedFilesList(config.base);
    const tracks = config.track;
    if (config.commands) {
        for (const commandName in config.commands) {
            if (config.commands.hasOwnProperty(commandName) &&
                (!filterCommand || commandName === filterCommand)) {
                const { track } = config.commands[commandName];
                if (track) {
                    tracks.push(...track);
                }
            }
        }
    }
    const projectTracks = tracks.filter((file) => file.includes("<project>"));
    const nonProjectTracks = tracks.filter((file) => !file.includes("<project>"));
    // TODO: .concat("bob.config.js");
    const nonProjectChanges = changedFiles.filter((file) => nonProjectTracks.includes(file));
    if (!nonProjectChanges.length) {
        changedFiles.forEach((file) => {
            for (const packageName in packages) {
                if (packages.hasOwnProperty(packageName)) {
                    const { location } = packages[packageName];
                    if (file.includes(location)) {
                        const tracks = projectTracks.map((path) => path.replace("<project>", location));
                        if (tracks.some((pattern) => minimatch_1.default(file, pattern))) {
                            packages[packageName].dirty = true;
                        }
                    }
                }
            }
        });
    }
    else {
        for (const packageName in packages) {
            if (packages.hasOwnProperty(packageName)) {
                packages[packageName].dirty = true;
            }
        }
    }
    const affected = Object.keys(packages).filter((name) => {
        const { dirty, dependencies } = packages[name];
        return dirty || dependencies.some((dep) => packages[dep].dirty);
    });
    return {
        affected,
        packages,
    };
}
exports.getAffectedPackages = getAffectedPackages;
function getChangedFilesList(against) {
    const revision = child_process_1.execSync(`git merge-base ${against} -- HEAD`, {
        encoding: "utf-8",
    });
    const cmd = child_process_1.execSync(`git diff --name-only ${revision}`, {
        encoding: "utf-8",
    });
    return cmd.split("\n").filter((file) => Boolean(file));
}
function getPackages(ignored = []) {
    const info = child_process_1.execSync("yarn workspaces info", {
        encoding: "utf-8",
    });
    const startsAt = info.indexOf("{");
    const endsAt = info.lastIndexOf("}");
    const workspaces = JSON.parse(info.substr(startsAt, endsAt - startsAt + 1));
    const packages = {};
    function collectDependencies(name, dependencies) {
        workspaces[name].workspaceDependencies.forEach((dep) => {
            if (!dependencies.includes(dep)) {
                dependencies.push(dep);
                collectDependencies(dep, dependencies);
            }
        });
    }
    for (const packageName in workspaces) {
        if (workspaces.hasOwnProperty(packageName) &&
            !ignored.includes(packageName)) {
            const { location } = workspaces[packageName];
            packages[packageName] = {
                location,
                dependencies: [],
                dirty: false,
            };
            collectDependencies(packageName, packages[packageName].dependencies);
            packages[packageName].dependencies = packages[packageName].dependencies.filter((name, i, all) => all.indexOf(name) === i);
        }
    }
    return packages;
}
exports.getPackages = getPackages;
