"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const fs_extra_1 = tslib_1.__importDefault(require("fs-extra"));
const path_1 = require("path");
const command_1 = require("../command");
const run_1 = require("./run");
const assert_1 = require("assert");
exports.validateCommand = command_1.createCommand((api) => {
    const { config, reporter } = api;
    return {
        command: "validate",
        describe: "Validate the monorepo",
        async handler() {
            const errors = [];
            const collectError = (err) => {
                errors.push(err);
            };
            const packages = run_1.getPackages(config.ignore);
            await Promise.all([
                validateRootPackage(collectError),
                validateGitIngore(collectError),
                validateRootTSConfig(collectError, packages),
                ...Object.keys(packages).map((name) => validatePackage({ name, packages, onError: collectError })),
            ]);
            errors.sort((a, b) => {
                if (a.message > b.message) {
                    return -1;
                }
                else {
                    return 1;
                }
            });
            if (errors.length === 0) {
                reporter.success("Everything looks okay!");
            }
            else {
                reporter.fatal("Please fix following errors:\n");
                errors.forEach((error) => {
                    reporter.error(error.message);
                });
                throw new Error("You monorepo setup is invalid");
            }
        },
    };
});
async function validateRootPackage(onError) {
    var _a;
    const pkg = await readJSON(path_1.resolve(process.cwd(), "package.json"));
    shouldEqual((_a = pkg === null || pkg === void 0 ? void 0 : pkg.scripts) === null || _a === void 0 ? void 0 : _a.build, "tsc --project tsconfig.json && bob build", `<root>/package.json: 'scripts.build' should equal 'tsc --project tsconfig.json && bob build'`, onError);
}
exports.validateRootPackage = validateRootPackage;
async function validateGitIngore(onError) {
    const gitignore = await fs_extra_1.default.readFile(path_1.resolve(process.cwd(), ".gitignore"), "utf-8");
    const lines = gitignore.split("\n");
    if (!lines.includes(".bob")) {
        onError(new Error("<root>/.gitignore should contain '.bob'"));
    }
    if (!lines.includes("dist")) {
        onError(new Error("<root>/.gitignore should contain 'dist'"));
    }
}
exports.validateGitIngore = validateGitIngore;
async function validateRootTSConfig(onError, packages) {
    const tsconfig = await readJSON(path_1.resolve(process.cwd(), "tsconfig.json"));
    shouldEqual(tsconfig.compilerOptions.incremental, true, "<root>/tsconfig.json: 'compilerOptions.incremental' !== true", onError);
    shouldEqual(tsconfig.compilerOptions.baseUrl, ".", "<root>/tsconfig.json: 'compilerOptions.baseUrl' !== '.'", onError);
    shouldEqual(tsconfig.compilerOptions.outDir, "dist", "<root>/tsconfig.json: 'compilerOptions.outDir' !== 'dist'", onError);
    shouldEqual(tsconfig.compilerOptions.module, "esnext", "<root>/tsconfig.json: 'compilerOptions.module' !== 'esnext'", onError);
    if (typeof tsconfig.compilerOptions.paths !== "object") {
        onError(new Error(`<root>/tsconfig.json: 'compilerOptions.paths' should exist`));
    }
    else {
        for (const name in packages) {
            if (packages.hasOwnProperty(name)) {
                const { location } = packages[name];
                if (!tsconfig.compilerOptions.paths[name] ||
                    !tsconfig.compilerOptions.paths[name].includes(path_1.join(location, "src/index.ts"))) {
                    onError(new Error(`<root>/tsconfig.json: 'compilerOptions.paths' should contain ${name} with ['${path_1.join(location, "src/index.ts")}']`));
                }
            }
        }
    }
    if (!tsconfig.include || !tsconfig.include.includes("packages")) {
        onError(new Error(`<root>/tsconfig.json: 'include' should contain 'package'`));
    }
}
exports.validateRootTSConfig = validateRootTSConfig;
async function validatePackage({ name, packages, onError, }) {
    var _a, _b;
    const pkg = await readJSON(path_1.resolve(process.cwd(), packages[name].location, "package.json"));
    shouldEqual(pkg.main, "dist/index.cjs.js", `${name}: 'main' should equal 'dist/index.cjs.js'`, onError);
    shouldEqual(pkg.module, "dist/index.esm.js", `${name}: 'module' should equal 'dist/index.esm.js'`, onError);
    shouldEqual(pkg.typings, "dist/index.d.ts", `${name}: 'typings' should equal 'dist/index.d.ts'`, onError);
    shouldEqual((_a = pkg.typescript) === null || _a === void 0 ? void 0 : _a.definition, "dist/index.d.ts", `${name}: 'typescript.definition' should equal 'dist/index.d.ts'`, onError);
    shouldEqual((_b = pkg.scripts) === null || _b === void 0 ? void 0 : _b.prepack, "bob prepack", `${name}: 'scripts.prepack' should equal 'bob prepack'`, onError);
    // TODO: if "bin" then "buildOptions.bin.$.input" equals "bin.$" (just file name, we shouldn't use directories here)
}
exports.validatePackage = validatePackage;
async function readJSON(filepath) {
    return fs_extra_1.default.readJSON(filepath, {
        encoding: "utf-8",
    });
}
function shouldEqual(value, expected, message, onError) {
    try {
        assert_1.equal(value, expected, message);
    }
    catch (error) {
        onError(error);
    }
}
